import re
from argparse import ArgumentParser
from csv import reader
from .pkg_resources import pkgfiles

from . import data
from .url import URL

DEFAULT_STATIONS_CSV = "stations.csv"
# dark cyan; RGB-format text: 72, 201, 176
ANSI_DK_CY = '\x1b[38;2;72;201;176m'
ANSI_YELO = '\x1b[38;2;244;208;63m'  # yellow
ANSI_DK_ORNG = '\x1b[38;2;220;118;51m'  # dark orange
ANSI_RESET = '\x1b[0m'  # reset text format to default


class StationListParseException(Exception):
    pass


class StationEntry:

    def __init__(self, name, description, url, is_video):
        self._name = name
        self._description = description
        self._url = URL(url)
        self._is_video = is_video

    @property
    def name(self):
        return self._name

    @property
    def description(self):
        return self._description

    @property
    def url(self):
        return str(self._url)

    @property
    def is_video(self):
        return self._is_video

    @property
    def port(self):
        return self._url.parsed.port

    def ansi_colorized(self):
        c2 = ANSI_YELO
        c3 = ANSI_DK_ORNG
        res = ANSI_RESET

        line = f'{c2}{self.name:}{res}  {c3}{self.description}{res}'
        return line


class StationList(dict):
    LIST_NAME = "Streaming Audio Channels"
    STATION_CSV = DEFAULT_STATIONS_CSV
    VIDEO_STREAMS = False

    def __init__(self, substring="", first_match=False, starting_idx=1):
        super().__init__()
        self._starting_idx = starting_idx
        if not substring:
            # first match doesn't make sense if substring is emtpy string
            first_match = False

        try_again = True
        while try_again:
            # If we're doing a substring match, we need to set the try_again flag
            # if not, we shouldn't try again
            try_again = len(substring) > 0
            self._populate_stations(substring, starting_idx, first_match)
            if len(self):
                # found at least one substring match (or no substring was provided)
                break
            # No substring matches, so eliminate the substring
            # TODO: Should we just raise an exception here?
            substring = ""
            continue

        self._exact_match = len(self) == 1

    @property
    def match(self):
        _match = None
        if self._exact_match:
            num = list(self.keys())[0]
            _match = self[num]
        return _match

    @property
    def last_idx(self):
        return self._starting_idx + len(self) - 1

    def _collapse_string(self, strarg):
        """
        Collapse all whitespace out of a string,
        and normalize to lower case
        """
        rex = re.compile(r'\W+')
        result = rex.sub('', strarg)
        result = result.lower()
        return result

    def has_station_num(self, station_num):
        return station_num in self.keys()

    def parse_csv_record(self, csv_record: list):
        try:
            name = csv_record[0]
            description = csv_record[1]
            url = csv_record[2]
        except IndexError as e:
            raise StationListParseException(
                "Unable to parse name, description, and URL from CSV record") from e
        return (name, description, url)

    def _populate_stations(self, substring, starting_idx, first_match: bool):
        substring = self._collapse_string(substring)
        with pkgfiles(data).joinpath(self.STATION_CSV).open("r") as _file:
            _reader = reader(_file)
            for number, csv_record in enumerate(_reader, starting_idx):
                try:
                    name, description, url = self.parse_csv_record(csv_record)
                except StationListParseException:
                    print(f"failed to parse {csv_record}")
                    continue
                if name.startswith("?"):
                    continue

                _name = self._collapse_string(name)
                if substring not in _name:
                    continue
                entry = StationEntry(
                    name, description, url, self.VIDEO_STREAMS)
                self[number] = entry
                if first_match:
                    break

    def ansi_colorized_line(self, number):
        entry: StationEntry = self[number]
        c1 = ANSI_DK_CY
        res = ANSI_RESET
        colorized_entry = entry.ansi_colorized()

        line = f'{c1}{number:>2}{res}  {colorized_entry}'
        return line

    def print_header(self):
        top_bottom = "=" * (len(self.LIST_NAME) + 4)
        # ============================
        # *                          *
        # * Streaming Audio Channels *
        # *                          *
        # ============================
        lines = [
            top_bottom,
            "*" + " " * (len(top_bottom) - 2) + "*",
            "* " + self.LIST_NAME + " *",
            "*" + " " * (len(top_bottom) - 2) + "*",
            top_bottom
        ]
        for line in lines:
            print(line)

    def print_menu(self):
        self.print_header()
        for snum in self.keys():
            line = self.ansi_colorized_line(snum)
            print(line)


def sl_parse_args():
    parser = ArgumentParser()
    subp = parser.add_subparsers()
    url_cmd = subp.add_parser("print-stations")
    url_cmd.add_argument("--urls", action="store_true")
    url_cmd.add_argument("--nonstandard-ports", action="store_true")

    parsed = parser.parse_args()
    return parsed


def station_list_main():
    config = sl_parse_args()
    station_list = StationList()
    if config.urls:
        ignore_ports = []
        if config.nonstandard_ports:
            ignore_ports = [80, 443, None]
        for _, station_entry in station_list.items():
            url = station_entry.url
            if station_entry.port not in ignore_ports:
                print(url)



import argparse

from .version import CSAbout


def vlc_parse_args():
    parser = argparse.ArgumentParser(description=str(CSAbout()))
    parser.add_argument(
        "station",
        help="Index or (partial) name of station to play",
        nargs="?",  # make this arg optional
    )
    parser.add_argument(
        "-f", "--first-match",
        help="Choose first partial station name match",
        action="store_true"
    )
    parser.add_argument(
        "-l", "--loop",
        help="Loop mode: Return to station menu when player terminates (q to quit completely)",
        action="store_true"
    )
    parser.add_argument(
        "--gui",
        help="Disable ncurses interface, run VLC in GUI mode",
        action="store_true",
    )
    parser.add_argument(
        "--write-shell-script",
        help="Write a shell script that sets up environment and executes vlc-radio",
        action="store_true"
    )
    parser.add_argument(
        "--version",
        help="Print version string and exit",
        action='version',
        version=str(CSAbout())
    )
    parser.add_argument(
        # just print the bare version string
        "--bare-version",
        help=argparse.SUPPRESS,
        action='version',
        version=CSAbout().version
    )

    parsed = parser.parse_args()
    return parsed



import glob
import os
import platform
import time
from pathlib import Path
from typing import List, Optional, Tuple

from . import logging
from .cmd import CMD
from .station_list import StationEntry


class VLCException(Exception):
    pass


class VLCLocator(CMD):
    VLC_PATH_ENV_VAR = "VLC_PATH"
    CMD_NAME = "which"
    ARGS = ["vlc"]

    def __init__(self):
        super().__init__(self.ARGS)
        self._location = self._locate()

    def _fix_exe_case(self, location):
        """
        Case insensitive filesystems will allow is to execute VLC or vlc,
        but things that match on process name, like AirFoil, look for the actual
        process name
        """
        if not location:
            raise Exception(f"no location: {location}")
        exe_upper = os.path.basename(location).upper()
        exe_lower = exe_upper.lower()
        dirname = os.path.dirname(location)
        dirglob = os.path.join(dirname, "*")
        items = glob.glob(dirglob)

        if location not in items:
            # No idea what we started with, so lets just reset to lower case
            location = os.path.join(dirname, exe_lower)

        if location not in items:
            # haven't found it yet, so lets try upper cse
            location = os.path.join(dirname, exe_upper)

        if location not in items:
            # nothing worked, so something has gone wrong, set to None
            location = None
        return location

    def _locate(self) -> str:
        loc = os.environ.get(self.VLC_PATH_ENV_VAR)
        if loc:
            if not os.path.exists(loc):
                loc = None
            else:
                loc = self._fix_exe_case(loc)
        if not loc:
            ret: int
            out_bytes, ret = self.run(capture_out=True, capture_err=True)
            out = out_bytes.decode("utf-8")
            out = out.rstrip()
            if ret == 0:
                loc = out
                loc = self._fix_exe_case(loc)
        if not loc:
            raise VLCException("Can't locate VLC")
        return loc

    @property
    def location(self):
        return self._location


class VLC(CMD):
    CMD_NAME = "vlc"
    PAUSE_SECS = 2.0

    def __init__(self, entry: StationEntry, ncurses: bool = True, vlc_path: Optional[str] = None, extra_args: List[str] = []):
        logger = logging.get_logger(__name__)
        self.entry = entry
        args = [entry.url]
        if entry.is_video:
            ncurses = False
        if ncurses:
            args.extend(["--intf", "ncurses"])
        else:
            args.extend(["--no-video-title-show", "--meta-title", entry.name])
        super().__init__(args, logger=logger)

        if not vlc_path:
            self._location = self._find_vlc()
        else:
            self._location = vlc_path

        self.argv[0] = self._location

    @property
    def location(self):
        return self._location

    def run(self) -> Tuple[bytes, int]:
        self._display_and_pause(self.PAUSE_SECS)
        run_tuple = super().run()
        self._check_iterm(run_tuple[1])
        return run_tuple

    def _display_and_pause(self, sec):
        print("")
        print("")
        print(f"Playing: {self.entry.ansi_colorized()}")
        print("")
        print("")
        time.sleep(sec)

    def _check_iterm(self, ret: int):
        terminfo_x_path = Path("~/.terminfo/x").expanduser()
        terminfo_s_path = Path("~/.terminfo/s").expanduser()

        if ret != 0 and "arm64" == platform.machine() and "Darwin" == platform.system():
            if not terminfo_x_path.exists() or not terminfo_s_path.exists():
                err_str = """
                It appears you're using arm64 macOS, where VLC has a bug. You may need to run:
                $ mkdir ~/.terminfo
                $ # for TERM=xterm-256color:
                $ ln -s /usr/share/terminfo/78 ~/.terminfo/x
                $ # for TERM=screen-256color:
                $ ln -s /usr/share/terminfo/73 ~/.terminfo/s
                """
                self.logger.error(f"{err_str}")

    def _find_vlc(self):
        locator = VLCLocator()
        loc = locator.location
        return loc



import sysconfig
from typing import Tuple, TypeAlias

ScriptPathHome: TypeAlias = str
ScriptPathUser: TypeAlias = str
ScriptPathPrefix: TypeAlias = str


def get_script_paths() -> Tuple[ScriptPathHome, ScriptPathUser, ScriptPathPrefix]:
    """
    Get Python script installation paths

    virtualenv example:
    - Home: /Users/zach/.virtualenvs/chill_streams/bin
    - User: /Users/zach/Library/Python/3.12/bin
    - prefix: /Users/zach/.virtualenvs/chill_streams/bin

    Non-virtualenv (homebrew python) example:
    - Home: /opt/homebrew/opt/python@3.12/Frameworks/Python.framework/Versions/3.12/bin
    - User: /Users/zach/Library/Python/3.12/bin
    - prefix: /opt/homebrew/bin

    Returns
    -------
    Tuple[ScriptPathHome, ScriptPathUser, ScriptPathPrefix]
        Tuple of script path strings for User, Home, and Prefix strings
    """

    script_paths = []
    scheme_keys = ["home", "user", "prefix"]
    # scheme_keys: List[Literal["user", "home", "prefix"]] = ["user", "home", "prefix"]
    for key in scheme_keys:
        scheme = sysconfig.get_preferred_scheme(key)  # type: ignore
        script_path = sysconfig.get_path("scripts", scheme)
        script_paths.append(script_path)

    return tuple(script_paths)



#!/usr/bin/env python
'''Play internet radio selection in VLC media player.
Shift + M --> meta-info; q --> quit'''

import argparse
import sys
from argparse import ArgumentParser
from typing import Optional

from . import logging
from .ascii_art import get_ascii_art
from .shell_script import VLCShellScript, VLCShellScriptException
from .station_list import StationEntry, StationList
from .version import CSAbout
from .video_streams import VideoStreamDependencyException, VideoStreamList
from .vlc import VLC


def vlc_parse_args():
    parser = ArgumentParser(description=str(CSAbout()))
    parser.add_argument(
        "station",
        help="Index or (partial) name of station to play",
        nargs="?",  # make this arg optional
    )
    parser.add_argument(
        "-f", "--first-match",
        help="Choose first partial station name match",
        action="store_true"
    )
    parser.add_argument(
        "-l", "--loop",
        help="Loop mode: Return to station menu when player terminates (q to quit completely)",
        action="store_true"
    )
    parser.add_argument(
        "--gui",
        help="Disable ncurses interface, run VLC in GUI mode",
        action="store_true",
    )
    parser.add_argument(
        "--write-shell-script",
        help="Write a shell script that sets up environment and executes vlc-radio",
        action="store_true"
    )
    parser.add_argument(
        "--version",
        help="Print version string and exit",
        action='version',
        version=str(CSAbout())
    )
    parser.add_argument(
        # just print the bare version string
        "--bare-version",
        help=argparse.SUPPRESS,
        action='version',
        version=CSAbout().version
    )
    parser.add_argument(
        "--debug", help="Enable debug logging", action='store_true')

    parser.add_argument(
        "--print-url", "-U", help="Print station URL to console and exit", action="store_true")

    parsed = parser.parse_args()
    return parsed


def station_selection(options):
    print(get_ascii_art())  # get color-schemed ASCII art heading
    '''Play selected internet radio station.'''
    station_text = ""
    station_num = -1
    ret = -1
    if options.station:
        try:
            station_num = int(options.station)
        except ValueError:
            station_text = options.station
    go_again = True
    while go_again:
        go_again = False
        station_list = StationList(
            substring=station_text, first_match=options.first_match)
        video_list: Optional[VideoStreamList] = None
        entry: Optional[StationEntry] = None
        if station_list.match:
            entry = station_list.match
        elif station_list.has_station_num(station_num):
            entry = station_list[station_num]
        else:
            idx = station_list.last_idx + 1
            try:
                # Only create the video stream list if we need to, since it
                # has to hit twitch.tv to query streams
                video_list = VideoStreamList(
                    substring=station_text, first_match=options.first_match, starting_idx=idx)
                if video_list.has_station_num(station_num):
                    entry = video_list[station_num]
                else:
                    entry = video_list.match
            except VideoStreamDependencyException:
                pass

        if not entry:
            station_list.print_menu()
            all_stations = dict(station_list)
            if video_list:
                all_stations.update(video_list)
                video_list.print_menu()
            inp = input('Enter item number [\'q\' to quit]: ')
            if inp in ['Q', 'q']:
                return 0
            station_num = int(inp)
            entry = all_stations[station_num]

        station_text = ""
        station_num = -1

        curses = not options.gui

        if entry:
            if options.print_url:
                ret = 0
                print("")
                print(f"{entry.ansi_colorized()}:\n    {entry.url}")
                print("")
            else:
                vlc = VLC(entry, ncurses=curses)
                _, ret = vlc.run()
                if ret != 0:
                    print(f"Failed to run {vlc.location}", file=sys.stderr)
                    break
                if options.loop:
                    go_again = True
        else:
            print("Station not found")

    return ret


def vlc_main():
    options = vlc_parse_args()
    if options.debug:
        logging.enable_debug_logging()

    if options.write_shell_script:
        try:
            vlcs = VLCShellScript()
            vlcs.write_script("~/.local/bin")
        except VLCShellScriptException as e:
            print(str(e))
            return -1
    else:
        station_selection(options)


if __name__ == '__main__':
    vlc_main()



try:
    import streamlink
except ImportError:
    streamlink = None

from .station_list import StationList, StationListParseException


class VideoStreamException(Exception):
    pass


class VideoStreamDependencyException(Exception):
    pass


class VideoStreamList(StationList):
    LIST_NAME = "Video Streams"
    STATION_CSV = "videostreams.csv"
    VIDEO_STREAMS = True

    def parse_csv_record(self, csv_record: list):
        name, description, url = super().parse_csv_record(csv_record)
        try:
            url = self.get_video_stream(url)
        except VideoStreamException as e:
            print(e)
            raise StationListParseException from e
        return (name, description, url)

    @staticmethod
    def get_video_stream(website_url):
        if not streamlink:
            raise VideoStreamDependencyException(
                "Can't extract video stream without 'streamlink' package")
        stream_url = None
        try:
            streams: dict = streamlink.streams(website_url)
        except (
                streamlink.NoPluginError,
                streamlink.PluginError) as e:
            raise VideoStreamException("Failed to extract video stream") from e
        try:
            stream_url = streams['best'].to_url()
        except KeyError:
            # 'best' wasn't available, so not sure what to do
            raise VideoStreamException(
                "Unable to find 'best' stream from streams dict")

        return stream_url



from .app import vlc_main
from .station_list import station_list_main


def main():
    try:
        return vlc_main()
    except KeyboardInterrupt:
        print("Keyboard interrupt. Quitting.")
        exit(0)


def sl_main():
    try:
        return station_list_main()
    except KeyboardInterrupt:
        exit(0)


if __name__ == "__main__":
    exit(main())



import urllib.parse as urllib_parse
from pysingleton import PySingleton

from . import data
from .pkg_resources import pkgfiles

"""
HTTPS upgrading
This module allows us to upgrade certain URLs to HTTPS if their domain is on the
list of domains known to support TLS.

This makes it possible to continue updating 'stations.csv' from upstream or other
sources but also upgrade to HTTPS where possible

TODO: handle URLs that are IP address only
TODO: handle URLs that are on ports other than 80 that support HTTPS on a port other than 443
"""


class HTTPSDomainList(list, metaclass=PySingleton):
    # TODO: Make this a singleton
    HTTPS_FILE = "https.txt"

    def __init__(self):
        super().__init__()
        with pkgfiles(data).joinpath(self.HTTPS_FILE).open("r") as _file:
            line: str
            for line in _file.readlines():
                if line.startswith("#"):
                    continue
                line = line.rstrip()
                if not line:
                    # skip empty lines
                    continue
                self.append(line)


class URL:
    def __init__(self, url_string):
        parsed = urllib_parse.urlparse(url_string)
        domain_list = HTTPSDomainList()
        self.parsed = self._https_upgrade(parsed, domain_list)

    def _https_upgrade(self, parsed_url, domain_list):
        netloc: str = parsed_url.netloc
        parts = netloc.split(":")
        hostname = parts.pop(0)
        port = parts.pop() if parts else None
        if port == "80":
            port = None
            parsed_url = parsed_url._replace(netloc=hostname)

        parts = hostname.split(".")
        domain = ".".join(parts[-2:])
        if domain in domain_list:
            if port is None:
                parsed_url = parsed_url._replace(scheme="https")
        return parsed_url

    def __str__(self):
        return urllib_parse.urlunparse(self.parsed)



from . import __summary__, __title__, __version__


class CSAbout:
    def __init__(self):
        self.version = __version__
        self.summary = __summary__
        self.title = __title__

    def __str__(self):
        return f"{self.title}: {self.summary} [version {self.version}]"



import logging.config
import logging

_log_level = logging.INFO


LOGGING = {
    'version': 1,
    'disable_existing_loggers': False,
    'formatters': {
        'simple': {
            'format': '[%(levelname)s] [%(name)s]: %(message)s'
        },
    },
    'handlers': {
        'console': {
            'level': 'DEBUG',
            'class': 'logging.StreamHandler',
            'formatter': 'simple'
        },
    },
    'loggers': {
        '': {
            'level': 'INFO',
            'propagate': True,
            'handlers': ['console'],
        },
    },
}

logging.config.dictConfig(LOGGING)


def enable_debug_logging():
    global _log_level
    _log_level = logging.DEBUG


def get_logger(logname):
    logger = logging.getLogger(logname)
    # logger.propagate = True

    logger.setLevel(_log_level)
    return logger



import re

from setuptools import find_packages, setup

about = {}
with open("chill_streams/__about__.py") as fp:
    exec(fp.read(), about)

with open("README.md", "r") as fp:
    long_description = fp.read()


GITHUB_URL = "https://github.com/zcutlip/chill_streams"
# image url: https://github.com/zcutlip/chill_streams/raw/main/images/radio-menu.png
# Image refs on PyPI should have absolute URLs to their home on github
# this awful regex looks for image refs:
#  ![any text](any/image/path), making sure there's no 'http[s]:'
# in the url part
# it then inserts hhttps://github.com/zcutlip/chill_streams/raw/main/
# between the '(' and the relative URL
# source: https://github.com/pypa/readme_renderer/issues/163#issuecomment-1679601106
long_description = re.sub(
    r"(!\[[^\]]+\]\()((?!https?:)[^\)]+)(\))",
    lambda m: m.group(1) + GITHUB_URL + "/raw/main/" +
    m.group(2) + m.group(3),
    long_description,
)

setup(
    name=about["__title__"],
    version=about["__version__"],
    description=about["__summary__"],
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Zachary Cutlip",
    author_email="uid000@gmail.com",
    url="https://github.com/zcutlip/chill_streams",
    license="MIT",
    packages=find_packages(),
    entry_points={
        "console_scripts": [
            "vlc-radio=chill_streams.cli:main",
            "sltool=chill_streams.cli:sl_main"
        ],
    },

    python_requires=">=3.10",
    install_requires=[
        "python-singleton-metaclasses"
    ],
    extras_require={
        'video-streams': ['streamlink']
    },
    package_data={"chill_streams": ["data/*"]},
)



import os


from . import data
from . import __version__ as chill_streams_version
from .pkg_resources import pkgfiles
from .script_path import get_script_paths
from .vlc import VLCLocator, VLCException


class VLCShellScriptException(Exception):
    pass


SCRIPT_TEMPLATE = "vlc-radio-template.sh"


class VLCShellScript:

    def __init__(self):
        self._vlc_path = self._locate_vlc()
        self._vlc_radio_path = self._locate_vlc_radio()

    @staticmethod
    def _locate_vlc():
        locator = VLCLocator()
        try:
            location = locator.location
        except VLCException as e:
            raise VLCShellScriptException(str(e)) from e
        return location

    def _locate_vlc_radio(self):
        script_path = None

        script_dirs = get_script_paths()
        for script_dir in script_dirs:
            script_path = os.path.join(script_dir, "vlc-radio")
            if os.path.exists(script_path):
                break
            script_path = None
        if not script_path:
            raise VLCShellScriptException("Can't locate script entrypoint")
        return script_path

    def _read_template(self):
        template = ""
        with pkgfiles(data).joinpath(SCRIPT_TEMPLATE).open("r") as _file:
            template = _file.read()
        return template

    def write_script(self, location):
        location = os.path.expanduser(location)
        location = os.path.normpath(location)
        os.makedirs(location, exist_ok=True, mode=0o755)
        script_path = os.path.join(location, "vlc-radio.sh")
        vlc_dir = os.path.dirname(self._vlc_path)
        radio_cmd = '$_VLC_RADIO_PATH "$@"'
        script_lines = [
            "",
            "# We need to export PATH before doing anything else",
            "# don't put any lines above this",
            f"export PATH=\"${{PATH}}\":{vlc_dir}",
            "",
            "# override default _VLC_RADIO_PATH",
            f"_VLC_RADIO_PATH={self._vlc_radio_path}",
            "# override default _CHILL_STREAMS_VERSION",
            f"_CHILL_STREAMS_VERSION={chill_streams_version}",
            "",
            "update_script && echo_sleep \"Restarting\" 1 && cleanup && exec \"$0\" \"$@\"",
            "",
            radio_cmd,
            ""
        ]
        script_template = self._read_template()
        script = script_template + "\n".join(script_lines)
        with open(os.open(script_path, os.O_CREAT | os.O_WRONLY | os.O_TRUNC, 0o755), "w") as f:
            f.write(script)



import shlex
import subprocess
from typing import List, Tuple

from .logging import get_logger


class CMD:
    CMD_NAME = None
    ARGS = []

    def __init__(self, argv: List[str] = [], logger=None):
        if logger is None:
            logger = get_logger(__name__)
        self.logger = logger
        if self.CMD_NAME is None:
            raise NotImplementedError("Override CMD_NAME with cmd to run")
        argv = [self.CMD_NAME] + argv
        self.argv = argv

    def run(self, capture_out=False, capture_err=False) -> Tuple[bytes, int]:
        out: bytes
        cmd_str = shlex.join(self.argv)
        self.logger.debug(f"About to run: {cmd_str}")
        returncode = 0
        kwargs = {}
        if capture_out:
            kwargs["stdout"] = subprocess.PIPE
        if capture_err:
            kwargs["stderr"] = subprocess.STDOUT

        try:

            p = subprocess.run(self.argv, check=True, **kwargs)
            out = p.stdout
        except subprocess.CalledProcessError as e:
            returncode = e.returncode
            out = e.stdout
        ret = (out, returncode)
        return ret
